"""buffer.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
import array


class BufferEncodeError(Exception):
    pass


class BufferDecodeError(Exception):
    pass


class Buffer(array.array):

    def __new__(cls, *args):
        return super().__new__(cls, 'B')

    def add_var_int32(self, val):
        if val & 127 == val:
            self.append(val)
            return

        if val >= 0x80000000 or val < -0x80000000:
            raise BufferEncodeError('int32 too big')

        if val < 0:
            val += 0x10000000000000000

        while True:
            bits = val & 127
            val >>= 7

            if val:
                bits |= 128

            self.append(bits)

            if not val:
                break

    def add_var_int64(self, val):
        if val >= 0x8000000000000000 or val < -0x8000000000000000:
            raise BufferEncodeError('int64 too big')

        if val < 0:
            val += 0x10000000000000000

        while True:
            bits = val & 127
            val >>= 7

            if val:
                bits |= 128

            self.append(bits)

            if not val:
                break

    def add_prefixed_string(self, val):
        assert isinstance(val, str), \
            'Expecting a str value but got {}'.format(type(val))
        self.add_var_int32(len(val))
        self.fromstring(val)



"""client_token.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
import os
import time
import json
import asyncio
import logging
import aiohttp
from urllib.parse import urlencode


AUTH_URL = 'https://accounts.google.com/o/oauth2/auth'
TOKEN_URL = 'https://accounts.google.com/o/oauth2/token'
REDIRECT_URI = 'urn:ietf:wg:oauth:2.0:oob'


class Token:

    _REQUIRED_TOKEN_FIELDS = {
        'refresh_token',
        'access_token',
        'scopes',
        'token_type',
        'expires_in',
        'timestamp'
    }

    def __init__(
            self,
            client_id,
            client_secret,
            token_file,
            scopes):
        self._lock = asyncio.Lock()
        self._client_id = client_id
        self._client_secret = client_secret
        self._token_file = token_file
        self._scopes = set(scopes)
        self._token = self._read_token_file()
        self._update_refresh_ts()

    def _update_refresh_ts(self):
        """Update _refresh_ts property which is set to the token initial
        timestamp plus half the expire time. The property is used to check
        if a token refresh is required.

        :return: None
        """
        self._refresh_ts = None if self._token is None else \
            self._token['timestamp'] + self._token['expires_in'] // 2

    def _read_token_file(self):
        """Read the specified token json file if available, checks the data and
        returns the token information.

        :return: None if the token file does not exist, a ValueError is raised
                 when the token file is not valid and a token dict is returned
                 if successful.
        """
        if not self._token_file.endswith('.json'):
            raise ValueError(
                'Invalid token file: {}, expecting a .json file.'
                .format(self._token_file))

        if os.path.isfile(self._token_file):
            with open(self._token_file, 'r') as f:
                token = json.load(f)

            missing = self._REQUIRED_TOKEN_FIELDS - set(token.keys())
            if missing:
                raise ValueError(
                    'Invalid token file: {} (fields are missing: {}).'
                    .format(self._token_file, ', '.join(missing))
                )

            scopes = set(token['scopes'])
            if scopes ^ self._scopes:
                raise ValueError(
                    'Invalid scopes defined in token file: {file}.\n'
                    'Expecting "{expecting}" but found "{found}".'.format(
                        file=self._token_file,
                        expecting=' '.join(self._scopes),
                        found=' '.join(scopes)
                    )
                )
            return token
        return None

    async def get(self):
        """Returns the access token. If _refresh_ts is passed, the token will
        be refreshed. A lock is used to prevent refreshing the token twice.

        :return: Access token (string)
        """
        async with self._lock:
            if self._refresh_ts < time.time():
                await self._refresh_token()
            return self._token['access_token']

    async def connect(self):
        """Connect to the google cloud. If no token file is found the user
        will be prompted to open a given link and copy/past an code. This
        usually is only required once and future access tokens will be
        received by using the refresh token.

        A lock is used to prevent connecting twice.

        :return: None
        """
        async with self._lock:
            if self._token is None:
                logging.info('Token is missing, request a new token.')
                self._token = await self._ask()
                self._save_token()
                self._update_refresh_ts()
                return

            assert self._refresh_ts is not None, \
                'When having a token we should also have a refresh_ts value.'

            if self._refresh_ts < time.time():
                await self._refresh_token()
            else:
                logging.info('Token is valid.')

    async def _refresh_token(self):
        logging.info(
            'Token has exceeded half of the expiration time, '
            'starting a token refresh.')

        data = {
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'grant_type': 'refresh_token',
            'refresh_token': self._token['refresh_token']
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(TOKEN_URL, data=data) as resp:
                content = await resp.text()
                content = json.loads(content)
                if resp.status == 200:
                    logging.info('Authentication successful')
                    self._token.update(content)
                    self._token['timestamp'] = int(time.time())
                    self._save_token()
                    self._update_refresh_ts()
                else:
                    raise ValueError(
                        'Got an error while asking for a token '
                        'refresh: {} ({})'
                        .format(
                            content.get('error', 'unknown'),
                            resp.status
                        ))

    def _save_token(self):
        path = os.path.dirname(self._token_file)
        if path and not os.path.exists(path):
            os.mkdir(path)
        with open(self._token_file, 'w') as f:
            json.dump(self._token, f)

    async def _ask(self):
        url = '{}?{}'.format(AUTH_URL, urlencode({
            'response_type': 'code',
            'approval_prompt': 'force',
            'access_type': 'offline',
            'scope': ' '.join(self._scopes),
            'client_id': self._client_id,
            'redirect_uri': REDIRECT_URI
        }))

        while True:
            code = input('''
Open a browser, copy/paste the following link and paste the code here:
{}
code: '''.format(url))

            if not code:
                continue

            data = {
                'code': code,
                'redirect_uri': REDIRECT_URI,
                'client_id': self._client_id,
                'client_secret': self._client_secret,
                'grant_type': 'authorization_code'
            }

            async with aiohttp.ClientSession() as session:
                async with session.post(TOKEN_URL, data=data) as resp:
                    content = await resp.text()
                    content = json.loads(content)
                    if resp.status == 200:
                        logging.info('Authentication successful')
                        content['timestamp'] = int(time.time())
                        content['scopes'] = list(self._scopes)
                        return content

                    print('Got an error: {} ({})'.format(
                        content.get('error', 'unknown'),
                        resp.status
                    ))



"""decoder.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
import base64
from .buffer import Buffer
from .buffer import BufferDecodeError


class Decoder(Buffer):

    _idx = None
    _end = None

    def __new__(cls, *args, ks=None):
        assert ks is not None, \
            'Key string is required, for example: Decoder(ks=<ket_string>)'

        decoder = super().__new__(cls)

        if isinstance(ks, str):
            ks = ks.encode('utf8')

        ks += b'=' * (4 - len(ks) % 4)
        ks = base64.b64decode(ks.replace(b'-', b'+').replace(b'_', b'/'))
        decoder.fromstring(ks)
        decoder._idx = 0
        decoder.set_end()
        return decoder

    def set_end(self, end=None):
        """Set a new _end relative to the current index or restore the original
        _end if no end is given."""
        self._end = len(self) if end is None else self._idx + end

    def __bool__(self):
        return self._idx < self._end

    def _get8(self):
        if not self:
            raise BufferDecodeError('truncated')
        c = self[self._idx]
        self._idx += 1
        return c

    def get_var_int32(self):
        b = self._get8()

        if not (b & 128):
            return b

        result = 0
        shift = 0

        while True:
            result |= ((b & 127) << shift)
            shift += 7

            if not (b & 128):
                if result >= 0x10000000000000000:
                    raise BufferDecodeError('corrupted')
                break

            if shift >= 64:
                raise BufferDecodeError('corrupted')

            b = self._get8()

        if result >= 0x8000000000000000:
            result -= 0x10000000000000000

        if result >= 0x80000000 or result < -0x80000000:
            raise BufferDecodeError('corrupted')

        return result

    def get_var_int64(self):
        result = 0
        shift = 0
        while True:
            if shift >= 64:
                raise BufferDecodeError('corrupted')

            b = self._get8()
            result |= (b & 127) << shift
            shift += 7

            if not (b & 128):
                if result >= 1 << 64:
                    raise BufferDecodeError('corrupted')

                break

        if result >= 1 << 63:
            result -= 1 << 64

        return result

    def get_prefixed_string(self):
        l = self.get_var_int32()
        if self._idx + l > len(self):
            raise BufferDecodeError('truncated')
        r = self[self._idx: self._idx + l]
        self._idx += l
        return r.tostring().decode('utf-8')



"""value.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from ...connector.entity import Entity
from ...connector.utils import value_to_dict


class Value:

    def __init__(self, default=None, required=True):
        self.default = default
        self.required = required
        self.name = None

    @property
    def ascending(self):
        return self.name, 'ASCENDING'

    @property
    def descending(self):
        return self.name, 'DESCENDING'

    def check_value(self, value):
        raise NotImplementedError()

    def get_value(self, model):
        return model.__dict__.get(self.name, None)

    def set_value(self, model, value):
        Entity.set_property(model, self.name, value)

    def _compare(self, other, op):
        self.check_value(other)
        return {
            'property': {
                'name': self.name
            },
            'value': value_to_dict(other),
            'op': op
        }

    def __eq__(self, other):
        return self._compare(other, 'EQUAL')

    def __ne__(self, other):
        raise Exception('Cannot use NOT EQUAL in a filter expression')

    def __lt__(self, other):
        return self._compare(other, 'LESS_THAN')

    def __le__(self, other):
        return self._compare(other, 'LESS_THAN_OR_EQUAL')

    def __gt__(self, other):
        return self._compare(other, 'GREATER_THAN')

    def __ge__(self, other):
        return self._compare(other, 'GREATER_THAN_OR_EQUAL')



"""__init__.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from .connector import GcdConnector, GcdServiceAccountConnector
from .client_token import Token
from .service_account_token import ServiceAccountToken



"""pathelement.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""


class TimestampValue:

    def __init__(self, timestamp_value):
        self._timestamp_value = timestamp_value

    def __str__(self):
        """Returns the formatted timestamp value."""
        return self._timestamp_value



"""anyvalue.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from .value import Value


class AnyValue(Value):

    def __init__(self, default=None, required=True, accept=None):
        """Initialize a mapped property.

        When 'accept' is None any type is accepted. A tuple can be
        used to force one of the specified types.

        :param default: default value, all types allowed
        :param required: boolean
        :param accept: None or tuple
        """
        self._accept = accept
        super().__init__(default=default, required=required)

    def check_value(self, value):
        if self._accept and not isinstance(value, self._accept):
            raise TypeError(
                'Received value for property {!r} is of invalid type: {!r}'
                .format(self.name, value.__class__.__name__))

    def set_value(self, model, value):
        self.check_value(value)
        super().set_value(model, value)



from urllib.parse import urlencode, quote_plus
from asyncio_extras.contextmanager import async_contextmanager
from asyncio_extras.asyncyield import yield_
import asyncio
import aiohttp
import datetime
import json
import jwt
import logging
import time
import typing

ScopeList = typing.List[str]
JWT_GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:jwt-bearer'
GCLOUD_TOKEN_DURATION = 3600
MISMATCH = "Project name passed to Token does not match service_file's " \
           "project_id."


@async_contextmanager
async def ensure_session(session):

    if session:
        await yield_(session)
    else:
        async with aiohttp.ClientSession() as session:
            await yield_(session)


class ServiceAccountToken():

    def __init__(self, project_id: str, service_file: str,
                 scopes: ScopeList, session: aiohttp.ClientSession=None):

        self.project_id = project_id

        self.service_data = None
        with open(service_file) as f:
            self.service_data = json.loads(f.read())

        # sanity check
        assert self.project_id == self.service_data['project_id'], MISMATCH

        self.scopes = list(scopes)

        self.session = session
        self.access_token = None
        self.access_token_duration = None
        self.access_token_acquired_at = None

        self.acquiring = None

    async def get(self):
        await self.ensure_token()
        return self.access_token

    async def connect(self):
        # Really just an alias of self.get
        token_ = await self.get()
        logging.info('Token is valid.')
        return token_

    async def ensure_token(self):
        if self.acquiring:
            await self.acquiring

        elif not self.access_token:
            self.acquiring = asyncio.ensure_future(
                self._acquire_access_token())
            await self.acquiring

        else:
            now = datetime.datetime.now()
            delta = (now - self.access_token_acquired_at).total_seconds()
            if delta > self.access_token_duration / 2:
                self.acquiring = asyncio.ensure_future(
                    self._acquire_access_token())

                await self.acquiring

    async def _acquire_access_token(self):
        data = await self._acquire_token()

        access_token = data['access_token']
        expires_in = data['expires_in']

        self.access_token = access_token
        self.access_token_duration = expires_in
        self.access_token_acquired_at = datetime.datetime.now()
        self.acquiring = None

        return True

    async def _acquire_token(self):
        assertion = self._generate_assertion()
        url = self.service_data['token_uri']

        headers = {'content-type': 'application/x-www-form-urlencoded'}
        payload = {
            'grant_type': JWT_GRANT_TYPE,
            'assertion': assertion
        }
        payload = urlencode(payload, quote_via=quote_plus)

        async with ensure_session(self.session) as s:
            response = await s.post(
                url,
                data=payload,
                headers=headers,
                timeout=60
            )
            json = await response.json()

        return json

    def _generate_assertion(self):
        payload = self._make_gcloud_oauth_body(
        )

        jwt_token = jwt.encode(
            payload,
            self.service_data['private_key'],
            algorithm='RS256'
        )

        return jwt_token

    def _make_gcloud_oauth_body(self):
        uri = self.service_data['token_uri']
        client_email = self.service_data['client_email']

        now = int(time.time())

        return {
            'aud': uri,
            'exp': now + GCLOUD_TOKEN_DURATION,
            'iat': now,
            'iss': client_email,
            'scope': ' '.join(self.scopes),
        }


async def _smoke_test(project, service_file, scopes=None):

    import aiohttp
    DEFAULT_SCOPES = {
        'https://www.googleapis.com/auth/datastore',
        'https://www.googleapis.com/auth/cloud-platform'
    }
    scopes = scopes or list(DEFAULT_SCOPES)

    with aiohttp.ClientSession() as session:

        token = ServiceAccountToken(
            project,
            service_file,
            session=session,
            scopes=scopes
        )

        result = await token.connect()

    print('success: {}'.format(result))


if __name__ == '__main__':
    import sys
    loop = asyncio.get_event_loop()
    loop.run_until_complete(_smoke_test(*sys.argv[1:]))



"""arrayvalue.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from .value import Value
from ..utils import ProtectedList


class ArrayValue(Value):

    def __init__(self, default=None, required=True, accept=None):
        """Initialize an array property.

        When 'accept' is None any type in the list is accepted. A tuple can be
        used to force each list item to be one of the specified types.

        For example:

        # my_array will only accept Key objects.
        my_array = ArrayValue(accept=(Key,))

        :param default: list or None
        :param required: boolean
        :param accept: None or tuple
        """
        self._accept = accept
        super().__init__(default=default, required=required)

    def check_value(self, value):
        if not isinstance(value, list):
            raise TypeError(
                'Expecting an value of type \'int\' for property {!r} '
                'but received type {!r}.'
                .format(self.name, value.__class__.__name__))

        if self._accept and \
                not all([isinstance(item, self._accept) for item in value]):
            raise TypeError(
                'At least one item in array property {!r} has an invalid type.'
                .format(self.name))

    def _protect(self, value):
        if self._accept and not isinstance(value, self._accept):
            raise TypeError('Invalid type {!r} for array property {!r}.'
                            .format(value.__class__.__name__, self.name))

    def set_value(self, model, value):
        self.check_value(value)
        super().set_value(model, ProtectedList(value, protect=self._protect))



from .value import Value
from ...connector.timestampvalue import TimestampValue
import re

RFC3339_RE = re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,9})?(?:(?:[\+\-]\d{2}:\d{2})|Z)$')  # nopep8


class DatetimeValue(Value):

    def check_value(self, value):
        if not isinstance(value, TimestampValue):
            raise TypeError(
                'Expecting an value of type \'TimestampValue\' for property '
                '{!r} but received type {!r}.'
                .format(self.name, value.__class__.__name__))

        if not RFC3339_RE.search(str(value)):
            raise TypeError(
                'Expecting a value of type \'str\' in RFC3339 datetime format '
                'for property {!r}.'
                .format(self.name))

    def set_value(self, model, value):
        if isinstance(value, str):
            value = TimestampValue(value)

        self.check_value(value)
        super().set_value(model, value)



"""jsonvalue.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
import json
from .value import Value


class JsonValue(Value):

    def __init__(self, default=None, required=True, accept=None):
        """Initialize an json property.

        When 'accept' is None any type is accepted as long as the value is
        JSON serializable. A tuple can be used to force each value to be one of
        the specified types.

        For example:

        # my_json_prop will only accept dictionary objects.
        my_json_prop = JsonValue(accept=(dict,))

        :param default: list or None
        :param required: boolean
        :param accept: None or tuple
        """
        self._accept = accept
        super().__init__(default=default, required=required)

    def check_value(self, value):
        if self._accept and not isinstance(value, self._accept):
            raise TypeError(
                'Received value for property {!r} is of invalid type: {!r}'
                .format(self.name, value.__class__.__name__))

    def set_value(self, model, value):
        self.check_value(value)
        try:
            data = json.dumps(value)
        except TypeError as e:
            raise TypeError('Value for property {!r} could not be parsed: {}'
                            .format(self.name, e))
        model.__dict__['__orig__{}'.format(self.name)] = value
        super().set_value(model, data)

    def get_value(self, model):
        key = '__orig__{}'.format(self.name)

        if key not in model.__dict__:
            try:
                model.__dict__[key] = json.loads(model.__dict__[self.name])
            except Exception as e:
                raise Exception(
                    'Error reading property {!r} '
                    '(see above exception for more info).'
                    .format(self.name)) from e

        return model.__dict__[key]



"""setup.py

Created on: May 19, 2017
   Authors: Jeroen van der Heijden <jeroen@transceptor.technology>
            jomido <https://github.com/jomido>
            egalpin <https://github.com/egalpin>

Upload to PyPI:

Build
  python setup.py sdist bdist_wheel

Upload to Test PyPI
  twine upload --repository-url https://test.pypi.org/legacy/ dist/*

Upload to PyPI
  twine upload dist/*
"""

import setuptools
from distutils.core import setup, Extension

VERSION = '0.11.0'

install_requires = [
    'aiohttp>=2',
    'PyJWT>=1',
    'cryptography>=1',
    'asyncio_extras>=1'
]

setup(
    name='aiogcd',
    packages=[
        'aiogcd',
        'aiogcd.connector',
        'aiogcd.orm',
        'aiogcd.orm.properties'],
    version=VERSION,
    description='Async Google Cloud Datastore API',
    author='Jeroen van der Heijden',
    author_email='jeroen@transceptor.technology',
    url='https://github.com/transceptor-technology/aiogcd',
    download_url='https://'
        'github.com/transceptor-technology/'
        'aiogcd/tarball/{}'.format(VERSION),
    keywords=['gcd', 'datastore', 'connector'],
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development'
    ],
)



"""connector.py

Created on: May 19, 2017
   Authors: Jeroen van der Heijden <jeroen@transceptor.technology>
            jomido <https://github.com/jomido>
"""
import json
import aiohttp
from .client_token import Token
from .service_account_token import ServiceAccountToken
from .entity import Entity
from .key import Key
from .utils import make_read_options

DEFAULT_SCOPES = {
    'https://www.googleapis.com/auth/datastore',
    'https://www.googleapis.com/auth/cloud-platform'
}

DATASTORE_URL = \
    'https://datastore.googleapis.com/v1/projects/{project_id}:{method}'

_MAX_LOOPS = 128


class GcdConnector:

    def __init__(
            self,
            project_id,
            client_id,
            client_secret,
            token_file,
            scopes=DEFAULT_SCOPES):

        self.project_id = project_id
        self._token = Token(
            client_id,
            client_secret,
            token_file,
            scopes)

        self._run_query_url = DATASTORE_URL.format(
            project_id=self.project_id,
            method='runQuery')

        self._commit_url = DATASTORE_URL.format(
            project_id=self.project_id,
            method='commit')

        self._lookup_url = DATASTORE_URL.format(
            project_id=self.project_id,
            method='lookup')

    async def connect(self):
        await self._token.connect()

    async def entities(self, entities):
        """Returns a tuple containing boolean values. Each boolean value is
        True in case of a successful mutation and False if not. The order of
        booleans is the same as the supplied tuple or list.

        Each entity must not already exist and each entity key will be updated
        with the new key in case no name/id was specified.

        :param entities: tuple or list with Entity objects
        :return: tuple containing boolean values
        """
        return self._commit_entities_or_keys(entities, 'insert')

    async def insert_entity(self, entity):
        """Returns True if successful or False if not. In case of False then
        most likely a conflict was detected.

        The entity must not already exist and the entity key will be updated
        with the new key in case no name/id was specified.

        :param entity: Entity object
        :return: Boolean
        """
        return (await self._commit_entities_or_keys([entity], 'insert'))[0]

    async def upsert_entities(self, entities):
        """Returns a tuple containing boolean values. Each boolean value is
        True in case of a successful mutation and False if not. The order of
        booleans is the same as the supplied tuple or list.

        Each entity may or may not already exist.

        :param entities: tuple or list with Entity objects
        :return: tuple containing boolean values
        """
        return self._commit_entities_or_keys(entities, 'upsert')

    async def upsert_entity(self, entity):
        """Returns True if successful or False if not. In case of False then
        most likely a conflict was detected.

        The entity may or may not already exist.

        :param entity: Entity object
        :return: Boolean
        """
        return (await self._commit_entities_or_keys([entity], 'upsert'))[0]

    async def update_entities(self, entities):
        """Returns a tuple containing boolean values. Each boolean value is
        True in case of a successful mutation and False if not. The order of
        booleans is the same as the supplied tuple or list.

        Each entity must already exist.

        :param entities: tuple or list with Entity objects
        :return: tuple containing boolean values
        """
        return self._commit_entities_or_keys(entities, 'update')

    async def update_entity(self, entity):
        """Returns True if successful or False if not. In case of False then
        most likely a conflict was detected.

        The entity must already exist.

        :param entity: Entity object
        :return: Boolean
        """
        return (await self._commit_entities_or_keys([entity], 'upsert'))[0]

    async def delete_keys(self, keys):
        """Returns a tuple containing boolean values. Each boolean value is
        True in case of a successful mutation and False if not. The order of
        booleans is the same as the supplied tuple or list.

        Each key may or may not already exist.

        :param keys: tuple or list with Key objects
        :return: tuple containing boolean values
        """
        return await self._commit_entities_or_keys(keys, 'delete')

    async def delete_key(self, key):
        """Returns True if successful or False if not. In case of False then
        most likely a conflict was detected.

        The key may or may not already exist.

        :param key: Key object
        :return: Boolean
        """
        return (await self._commit_entities_or_keys([key], 'delete'))[0]

    async def commit(self, mutations):
        """Commit mutations.

        The only supported commit mode is NON_TRANSACTIONAL.

        See the link below for information for a description of a mutation:

        https://cloud.google.com/datastore/docs/reference/
                rest/v1/projects/commit#Mutation

        :param mutations: List or tuple with mutations
        :return: tuple containing mutation results
        """
        data = {
            'mode': 'NON_TRANSACTIONAL',
            'mutations': mutations
        }
        async with aiohttp.ClientSession() as session:

            async with session.post(
                    self._commit_url,
                    data=json.dumps(data),
                    headers=await self._get_headers()) as resp:

                content = await resp.json()

                if resp.status == 200:
                    return tuple(content.get('mutationResults', tuple()))

                raise ValueError(
                        'Error while committing to the datastore: {} ({})'
                        .format(
                            content.get('error', 'unknown'),
                            resp.status
                        ))

    async def run_query(self, data):
        """Return entities by given query data.

        :param data: see the following link for the data format:
            https://cloud.google.com/datastore/docs/reference/rest/
                v1/projects/runQuery
        :return: list containing Entity objects.
        """
        results, _ = self._run_query(data)
        return results

    async def _run_query(self, data):
        results = []
        cursor = None
        while True:
            async with aiohttp.ClientSession() as session:

                if cursor is not None:
                    data['query']['startCursor'] = cursor

                async with session.post(
                        self._run_query_url,
                        data=json.dumps(data),
                        headers=await self._get_headers()) as resp:

                    content = await resp.json()

                    if resp.status == 200:

                        entity_results = \
                            content['batch'].get('entityResults', [])

                        results.extend(entity_results)

                        more_results = content['batch']['moreResults']
                        cursor = content['batch']['endCursor']

                        if more_results in (
                                'NO_MORE_RESULTS',
                                'MORE_RESULTS_AFTER_LIMIT',
                                'MORE_RESULTS_AFTER_CURSOR'):
                            break

                        if more_results == 'NOT_FINISHED':
                            continue

                        raise ValueError(
                            'Unexpected value for "moreResults": {}'
                            .format(more_results))

                    raise ValueError(
                        'Error while query the datastore: {} ({})'
                        .format(
                            content.get('error', 'unknown'),
                            resp.status
                        )
                    )

        return results, cursor

    async def _get_entities_cursor(self, data):
        results, cursor = await self._run_query(data)
        return [Entity(result['entity']) for result in results], cursor

    async def get_entities(self, data):
        """Return entities by given query data.

        :param data: see the following link for the data format:
            https://cloud.google.com/datastore/docs/reference/rest/
                v1/projects/runQuery
        :return: list containing Entity objects.
        """
        results, _ = await self._run_query(data)
        return [Entity(result['entity']) for result in results]

    async def get_keys(self, data):
        data['query']['projection'] = [{'property': {'name': '__key__'}}]
        results, _ = await self._run_query(data)
        return [Key(result['entity']['key']) for result in results]

    async def get_entity(self, data):
        """Return an entity object by given query data.

        :param data: see the following link for the data format:
            https://cloud.google.com/datastore/docs/reference/rest/
                v1/projects/runQuery
        :return: Entity object or None in case no entity was found.
        """
        data['query']['limit'] = 1
        result = await self.get_entities(data)
        return result[0] if result else None

    async def get_key(self, data):
        data['query']['limit'] = 1
        result = await self.get_keys(data)
        return result[0] if result else None

    async def get_entities_by_kind(self, kind):
        """Returns entities by kind.

        This is a shortcut for:
        get_entities({'query': {'kind': [{'name': kind}]}})
        """
        data = {'query': {'kind': [{'name': kind}]}}
        return await self.get_entities(data)

    async def get_entities_by_keys(self, keys, missing=None, deferred=None,
                                   eventual=False):
        """Returns entity objects for the given keys or None in case no
        entity is found.

        :param keys: list of Key objects
        :return: list of Entity objects or None.
        """
        read_options = make_read_options(eventual=eventual)
        data = lambda: json.dumps({
            'readOptions': read_options,
            'keys': [k.get_dict() for k in keys],
        })

        if missing is not None and missing != []:
            raise ValueError('missing must be None or an empty list')

        if deferred is not None and deferred != []:
            raise ValueError('deferred must be None or an empty list')

        attempts = 0
        entities = []
        while keys and attempts < _MAX_LOOPS:
            attempts += 1
            async with aiohttp.ClientSession() as session:
                async with session.post(
                        self._lookup_url,
                        data=data(),
                        headers=await self._get_headers()) as resp:

                    content = await resp.json()
                    entities.extend(Entity(result['entity']) for result in
                                    content.get('found', []))

                    if missing is not None:
                        missing.extend(result['entity'] for result in
                                       content.get('missing', []))

                    deferred_keys = [Key(result) for result in
                                     content.get('deferred', [])]

                    if deferred is not None:
                        deferred.extend(deferred_keys)
                        break

                    keys = deferred_keys

        if entities:
            return entities

    async def get_entity_by_key(self, key, missing=None, deferred=None,
                                eventual=False):
        """Returns an entity object for the given key or None in case no
        entity is found.

        :param key: Key object
        :return: Entity object or None.
        """
        entity = await self.get_entities_by_keys([key], missing, deferred,
                                                 eventual)
        if entity:
            return entity[0]

    async def _get_headers(self):
        token = await self._token.get()
        return {
            'Authorization': 'Bearer {}'.format(token),
            'Content-Type': 'application/json'
        }

    @staticmethod
    def _check_mutation_result(entity_or_key, mutation_result):
        if 'key' in mutation_result:
            # The automatically allocated key.
            # Set only when the mutation allocated a key.
            entity_or_key.key = Key(mutation_result['key'])

        return not mutation_result.get('conflictDetected', False)

    async def _commit_entities_or_keys(self, entities_or_keys, method):
        mutations = [
            {method: entity_or_key.get_dict()}
            for entity_or_key in entities_or_keys]

        mutations_results = await self.commit(mutations)

        return tuple(
            self._check_mutation_result(entity_or_key, mutation_result)
            for entity_or_key, mutation_result
            in zip(entities_or_keys, mutations_results))


class GcdServiceAccountConnector(GcdConnector):
    def __init__(
            self,
            project_id,
            service_file,
            session=None,
            scopes=None):

        scopes = scopes or list(DEFAULT_SCOPES)
        self.project_id = project_id
        self._token = ServiceAccountToken(project_id, service_file, scopes,
                                          session)

        self._run_query_url = DATASTORE_URL.format(
            project_id=self.project_id,
            method='runQuery')

        self._commit_url = DATASTORE_URL.format(
            project_id=self.project_id,
            method='commit')

        self._lookup_url = DATASTORE_URL.format(
            project_id=self.project_id,
            method='lookup')



"""filter.py

Created on: May 19, 2017
    Author: Jeroen van der Heijden <jeroen@transceptor.technology>
"""
from ..connector.key import Key
from ..connector import GcdConnector


class Filter(dict):

    def __init__(self, model, *filters, has_ancestor=None, key=None):
        self._model = model
        self._cursor = None
        filters = list(filters)

        if has_ancestor is not None:
            assert isinstance(has_ancestor, Key), \
                'Keyword argument \'has_ancestor\' should be of type ' \
                '\'Key\' but found type {!r}' \
                .format(has_ancestor.__class__.__name__)

            filters.append({
                'property': {'name': '__key__'},
                'value': {'keyValue': has_ancestor.get_dict()},
                'op': 'HAS_ANCESTOR'})

        if key is not None:
            assert isinstance(key, Key), \
                'Keyword argument \'key\' should be of type \'Key\' ' \
                'but found type {!r}' \
                .format(key.__class__.__name__)
            filters.append({
                'property': {'name': '__key__'},
                'value': {'keyValue': key.get_dict()},
                'op': 'EQUAL'})

        filter_dict = {'query': {'kind': [{'name': self._model.get_kind()}]}}

        if len(filters) == 1:
            filter_dict['query']['filter'] = {
                'propertyFilter': filters[0]
            }
        elif len(filters) > 1:
            filter_dict['query']['filter'] = {
                'compositeFilter': {
                    'op': 'AND',
                    'filters': [{'propertyFilter': f} for f in filters]
                }
            }

        super().__init__(**filter_dict)

    def _set_start_cursor(self, start_cursor):
        if start_cursor:
            if not isinstance(start_cursor, str):
                raise TypeError(
                    'start_cursor is expected to be str, {} passed'.format(
                        type(start_cursor)))
            self['query']['startCursor'] = start_cursor

    def _set_offset(self, offset):
        if offset:
            if not isinstance(offset, int):
                raise TypeError(
                    'offset is expected to be int, {} passed'.format(
                        type(offset)))

            self['query']['offset'] = offset

    def _set_limit(self, limit):
        if limit:
            if not isinstance(limit, int):
                raise TypeError(
                    'limit is expected to be int, {} passed'.format(
                        type(limit)))

            self['query']['limit'] = limit

    @property
    def cursor(self):
        return self._cursor

    def order_by(self, *order):
        self['query']['order'] = [
            {
                'property': {'name': p[0]},
                'direction': p[1]
            } if isinstance(p, tuple) else {
                'property': {'name': p.name},
                'direction': 'DIRECTION_UNSPECIFIED'
            }
            for p in order
        ]
        return self

    def limit(self, limit, start_cursor=None):
        self._set_limit(limit)
        self._set_start_cursor(start_cursor)
        return self

    async def get_entity(self, gcd: GcdConnector):
        """Return a GcdModel instance from the supplied filter.

        :param gcd: GcdConnector instance.
        :return: GcdModel object or None in case no entity was found.
        """
        entity = await gcd.get_entity(self)
        return None if entity is None else self._model(entity)

    async def get_entities(
            self, gcd: GcdConnector, offset=None, limit=None) -> list:
        """Returns a list containing GcdModel instances from the supplied
        filter.

        :param gcd: GcdConnector instance.
        :param offset: integer to specify how many rows to skip
        :param limit: integer to specify max number of rows to return
        :return: list containing GcdModel objects.
        """
        self._set_offset(offset)
        self._set_limit(limit)
        entities, cursor = await gcd._get_entities_cursor(self)
        self._cursor = cursor
        return [self._model(ent) for ent in entities]

    async def get_key(self, gcd: GcdConnector):
        """Return a Gcd key from the supplied filter.

        :param gcd: GcdConnector instance.
        :return: GcdModel key or None in case no entity was found.
        """
        return await gcd.get_key(self)

    async def get_keys(
            self, gcd: GcdConnector, offset=None, limit=None) -> list:
        """Returns a list containing Gcd keys from the supplied filter.

        :param gcd: GcdConnector instance.
        :param offset: integer to specify how many keys to skip
        :param limit: integer to specify max number of keys to return
        :return: list containing Gcd key objects.
        """
        self._set_offset(offset)
        self._set_limit(limit)
        return await gcd.get_keys(self)

    def set_offset_limit(self, offset, limit):
        """Set offset and limit for Filter query.
        :param offset: can be int or None(to avoid setting offset)
        :param limit: can be int or None(to avoid setting limit)
        :return: True: always returns True
        """
        self._set_offset(offset)
        self._set_limit(limit)
        return True
